package ca.recrown.islandsurvivalcraft.pathfinding;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.Objects;
import java.util.PriorityQueue;
import java.util.Queue;

import org.apache.commons.lang.NullArgumentException;

public class DepthFirstSearch {
    private Queue<Node> queue;
    private CoordinateValidatable validatable;
    private Node startNode;
    private Node endNode;
    private HashSet<Node> checkedNodes;

    public DepthFirstSearch() {
        checkedNodes = new HashSet<>();
    }

    public DepthFirstSearch(CoordinateValidatable validatable) {
        this();
        this.validatable = validatable;
    }

    public void setValidatable(CoordinateValidatable validatable) {
        this.validatable = validatable;
    } 

    public void setStartPosition(int x, int y) {
        startNode = new Node(null, x, y);
    }

    public void setEndPosition(int x, int y) {
        endNode = new Node(null, x, y);
    }

    public int getEndX() {
        return endNode.x;
    }

    public int getEndY() {
        return endNode.y;
    }

    public void deleteTree() {
        startNode = null;
        endNode = null;
    }

    public boolean buildPathToEndNode() {
        if (validatable == null) throw new IllegalStateException("Need to set a validator.");
        if (startNode == null) throw new IllegalStateException("Need to set the starting position.");

        queue = new PriorityQueue<>();
        checkedNodes.clear();
        Node begin = startNode;
        queue.add(begin);
        while (!queue.isEmpty()) {
            Node n = queue.poll();
            if (!n.equals(endNode)) {
                n.child[0] = new Node(n, n.x + 1, n.y);
                n.child[1] = new Node(n, n.x - 1, n.y);
                n.child[2] = new Node(n, n.x, n.y + 1);
                n.child[3] = new Node(n, n.x, n.y - 1);
            } else {
                n.shallowCopyTo(endNode);
                return true;
            }
            if (checkedNodes.add(n)) {
                for (int i = 0; i < n.child.length; i++) {
                    Node child = n.child[i];
                    if (validatable.validate(child.x, child.y)) {
                        queue.add(child);
                    }
                }
            }
        }
        return false;
    }

    /**
     * Finds the end node.
     * Will not set end node if one wasn't found,
     * therefore, previous end node data is kept.
     * @param targetValidator
     * @return
     */
    public boolean findTarget(CoordinateTargetValidatable targetValidator) {
        if (validatable == null) throw new IllegalStateException("Need to set a validator.");
        if (targetValidator == null) throw new NullArgumentException("targetValidator");
        if (startNode == null) throw new IllegalStateException("Need to set the starting position.");

        queue = new LinkedList<>();
        checkedNodes.clear();
        Node begin = startNode;
        queue.add(begin);
        while (!queue.isEmpty()) {
            Node n = queue.poll();
            if (!targetValidator.isCoordinateTarget(n.x, n.y)) {
                n.child[0] = new Node(n, n.x + 1, n.y);
                n.child[1] = new Node(n, n.x - 1, n.y);
                n.child[2] = new Node(n, n.x, n.y + 1);
                n.child[3] = new Node(n, n.x, n.y - 1);
            } else {
                endNode = n;
                return true;
            }
            if (checkedNodes.add(n)) {
                for (int i = 0; i < n.child.length; i++) {
                    Node child = n.child[i];
                    if (validatable.validate(child.x, child.y)) {
                        queue.add(child);
                    }
                }
            }
        }
        return false;
    }

    private class Node implements Comparable<Node> {
        public Node[] child;
        public Node parent;
        public int x;
        public int y;

        public Node(Node parent, int x, int y) {
            this.parent = parent;
            child = new Node[4];
            this.x = x;
            this.y = y;
        }

        @Override
        public int compareTo(Node o) {
            return Math.round(distanceToGoal(endNode) - o.distanceToGoal(endNode));
        }
        

        public float distanceToGoal(Node goal) {
            float distanceX = goal.x - this.x;
            float distanceY = goal.y - this.y;
            float distance = (float) Math.sqrt(distanceX * distanceX + distanceY * distanceY);
            return distance;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj instanceof Node) {
                Node comp = (Node) obj;
                if (comp.x == this.x && comp.y == this.y) {
                    return true;
                }
            }
            return false;
        }

        @Override
        public int hashCode() {
            return Objects.hash(x, y);
        }

        public void shallowCopyTo(Node node) {
            node.child = this.child;
            node.parent = this.parent;
            node.x = this.x;
            node.y = this.y;
        }
    }
}